from omnitools import overloaded, overload_condition
from typing import overload
from .utils import Mega
import threadwrapper
import shadowwalker
import traceback
import threading
import sqlq
import time
import json
import os


class MegaDownloader:
    clashes = None
    worker = None
    terminate = False
    terminated = False

    def __init__(
            self,
            quick_proxies: bool = False,
            proxies: list = None,
            db_i: int = 0,
            sqlq_server: list = None,
            clash_quiet: bool = True
    ):
        self.db_i = db_i
        self.quick_proxies = quick_proxies
        if self.quick_proxies:
            self.clashes = [shadowwalker.ShadowWalker(
                bench_file=os.path.join(shadowwalker.pkg_data_dir, "bench_mega.json")
            )]
            fast_proxy = self.clashes[-1].fast_proxy
            self.proxies = [self.clashes[-1].proxy]
            for i in range(1, len(fast_proxy)-1):
                self.clashes.append(self.clashes[i-1].clone())
                self.proxies.append(self.clashes[i].proxy)
            self.start_clash(clash_quiet)
        else:
            self.proxies = proxies or [None]
        self.login()
        if not sqlq_server:
            self.sqlqueue = sqlq.SqlQueueU(server=True, host="0.0.0.0", db="db_{}.db".format(db_i))
            self.sqlqueue.sql('''
                PRAGMA journal_mode = WAL;
            ''')
            self.sqlqueue.sql(
                '''
                CREATE TABLE IF NOT EXISTS `queue` (
                    `file` INTEGER,
                    `data` TEXT
                );
                CREATE TABLE IF NOT EXISTS `failed` (
                    `id` TEXT
                );
                CREATE TABLE IF NOT EXISTS `done` (
                    `id` TEXT
                );
                '''
            )
        else:
            self.sqlqueue = sqlq.SqlQueueU(host=sqlq_server[0], db_port=sqlq_server[1])
        self.sqlqueue.commit()

    def start_clash(self, quiet=True):
        for clash in self.clashes:
            p = threading.Thread(target=lambda: clash.start(quiet))
            p.daemon = True
            p.start()

    def login(self, *args, **kwargs):
        self.client = Mega()
        self.client.login(*args, **kwargs)
        print("logged in")

    @staticmethod
    def parse_args_uni(args, kwargs, i, key, default = Exception):
        try:
            return kwargs[key]
        except:
            try:
                return args[i]
            except:
                if default is Exception:
                    raise Exception("no args with ({}, {}) supplied".format(i, key))
                return default

    @overload
    def download(self, url, dest_path=None, serial_download=True, parallel_download=False, parallel_threads=2**3):
        ...

    @overload
    def download(self, url, dest_path=None, dest_filename=None, serial_download=True, parallel_download=False, parallel_threads=2**3):
        ...

    def download_folder(self, url, dest_path=None, serial_download=True, parallel_download=False, parallel_threads=2**3):
        data = []
        id, _ = self.client.parse_folder_url(url)
        files = self.client.get_files_in_folder(url, True)
        for i, file in enumerate(files):
            data.append((
                0,
                json.dumps([
                    url,
                    files[i],
                    len(files),
                    dest_path,
                    i,
                    serial_download,
                    parallel_download,
                    parallel_threads,
                    id
                ])
            ))
        if data:
            self.sqlqueue.sql(
                '''
                INSERT INTO `queue` VALUES (?, ?);
                ''',
                tuple(data)
            )
            self.sqlqueue.commit()

    def download_file(self, url, dest_path=None, dest_filename=None, serial_download=True, parallel_download=False, parallel_threads=2**3):
        id, _ = self.client._parse_url(url).split("!")
        try:
            self.client.get_file_data(id)
        except:
            traceback.print_exc()
            print("skipped url", url, "may not be available")
        data = (
            1,
            json.dumps([
                url,
                dest_path,
                dest_filename,
                serial_download,
                parallel_download,
                parallel_threads,
                id
            ])
        )
        if data:
            self.sqlqueue.sql(
                '''
                INSERT INTO `queue` VALUES (?, ?);
                ''',
                tuple(data)
            )
            self.sqlqueue.commit()

    @overloaded(
        overload_condition(
            args_i=0,
            kwargs_k="url",
            condition=lambda x: not ("/folder/" in x or "/#F!" in x),
            func=download_file
        ),
        overload_condition(
            args_i=0,
            kwargs_k="url",
            condition=lambda x: "/folder/" in x or "/#F!" in x,
            func=download_folder
        ),
    )
    def download(self, *args, **kwargs):
        ...

    def downloader(self, quiet=True, once=False, stop_on_exception=False):
        while not self.terminate:
            r = self.sqlqueue.sql(
                '''
                SELECT ROWID, *
                FROM `queue`
                ORDER BY ROWID ASC
                LIMIT 1;
                ''',
                (),
                "list"
            )
            if not r:
                time.sleep(1)
                continue
            rowid = r[0][0]
            args = json.loads(r[0][2])
            id = args.pop(-1)
            if r[0][1]:
                func = self.client.download_url
            else:
                func = self.client.download_file_in_folder
            # if self.quick_proxies:
            #     self.stop_clash()
            #     time.sleep(1)
            #     self.start_clash(quiet)
            try:
                self.sqlqueue.sql(
                    '''
                    DELETE FROM `done`
                    WHERE `id` = ?;
                    ''',
                    (id,)
                )
                self.sqlqueue.commit()
                func(*args, verbose=not quiet, proxies=self.proxies, stop_on_exception=stop_on_exception)
                self.sqlqueue.sql(
                    '''
                    DELETE FROM `queue`
                    WHERE ROWID = ?;
                    ''',
                    (rowid,)
                )
                self.sqlqueue.commit()
                self.sqlqueue.sql(
                    '''
                    INSERT INTO `done` VALUES (?);
                    ''',
                    (id,)
                )
                self.sqlqueue.commit()
                if once:
                    raise Exception("run once", once)
            except Exception as e:
                if "not found" in str(e):
                    self.sqlqueue.sql(
                        '''
                        DELETE FROM `queue`
                        WHERE ROWID = ?;
                        ''',
                        (rowid,)
                    )
                    self.sqlqueue.commit()
                    self.sqlqueue.sql(
                        '''
                        INSERT INTO `failed` VALUES (?);
                        ''',
                        (id,)
                    )
                    self.sqlqueue.commit()
                elif "transfer quota exceed" in str(e):
                    print("downloader has stopped due to exceed transfer quota for {}".format("proxy IP {}".format(e.args[1]) if e.args[1] else "your IP"))
                    self.terminate = True
                else:
                    traceback.print_exc()
                    print("downloader has stopped due to pending exception. see above traceback log.")
                    self.terminate = True
        self.terminated = True

    def start(self, quiet=True, daemon=True, once=False, stop_on_exception=False):
        t = lambda: self.downloader(quiet, once, stop_on_exception)
        if daemon:
            self.worker = threading.Thread(target=t)
            self.worker.daemon = daemon
            self.worker.start()
        else:
            t()
            self.terminate = False
            self.terminated = False

    def stop(self):
        if self.worker:
            self.terminate = True
            while not self.terminated:
                time.sleep(1)
        self._stop()

    def _stop(self):
        self.sqlqueue.commit()
        self.sqlqueue.stop()
        self.stop_clash()

    def stop_clash(self):
        if self.clashes:
            tw = threadwrapper.ThreadWrapper(threading.Semaphore(2**8))
            for clash in self.clashes:
                tw.add(job=lambda: clash.stop())
            tw.wait()


